from typing import Literal

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName


class SplitStep(BaseStep):
    name: Literal["split"] = "split"
    column: ColumnName
    delimiter: str
    # at least one col to keep
    number_cols_to_keep: int = Field(gt=0)


class SplitStepWithVariable(SplitStep, StepWithVariablesMixin):
    ...



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import SubstringStep


def translate_substring(step: SubstringStep) -> list[MongoStep]:
    pos_start_index = (
        step.start_index - 1
        if step.start_index > 0
        else {"$add": [{"$strLenCP": f"${step.column}"}, step.start_index]}
    )

    pos_end_index = (
        step.end_index - 1
        if step.end_index > 0
        else {"$add": [{"$strLenCP": f"${step.column}"}, step.end_index]}
    )

    length_to_keep = {
        "$add": [
            {
                "$subtract": [pos_end_index, pos_start_index],
            },
            1,
        ],
    }

    substr_mongo = {"$substrCP": [f"${step.column}", pos_start_index, length_to_keep]}

    return [
        {
            "$addFields": {
                (
                    step.new_column_name if step.new_column_name else f"{step.column}_SUBSTR"
                ): substr_mongo
            }
        }
    ]



from typing import Literal

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.combination import Reference


class DomainStep(BaseStep):
    name: Literal["domain"] = "domain"
    domain: str | Reference



from typing import Any, Literal

from pydantic import BaseConfig, BaseModel, Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName

from ..conditions import Condition


class IfThenElse(BaseModel):
    class Config(BaseConfig):
        allow_population_by_field_name = True

    condition: Condition = Field(alias="if")
    then: Any
    else_value: "IfThenElse" | Any = Field(alias="else")


IfThenElse.update_forward_refs()


class IfthenelseStep(BaseStep, IfThenElse):
    name: Literal["ifthenelse"] = "ifthenelse"
    new_column: ColumnName = Field(alias="newColumn")

    class Config(BaseConfig):
        allow_population_by_field_name = True


class IfThenElseStepWithVariables(IfthenelseStep, StepWithVariablesMixin):
    class Config(BaseConfig):
        allow_population_by_field_name = True



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import AbsoluteValueStep


def translate_absolutevalue(step: AbsoluteValueStep) -> list[MongoStep]:
    return [{"$addFields": {step.new_column: {"$abs": f"${step.column}"}}}]



from typing import Literal

from pydantic import Field, validator

from weaverbird.pipeline.steps.utils.base import BaseStep, to_camelcase


class HierarchyStep(BaseStep):
    class Config:
        alias_generator = to_camelcase

    name: Literal["hierarchy"] = "hierarchy"
    hierarchy_level_column: str = Field(default="hierarchy_level")
    hierarchy: list[str]
    include_nulls: bool = Field(default=False)

    @validator("hierarchy")
    def _ensure_hierarchy(cls, values: list[str]) -> list[str]:
        assert len(values) > 0, "at least one value must be specified"
        return values



from pandas import DataFrame

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import ConvertStep

from .utils.cast import cast_to_bool, cast_to_datetime, cast_to_float, cast_to_int, cast_to_str

CAST_FUNCTIONS = {
    "integer": cast_to_int,
    "float": cast_to_float,
    "text": cast_to_str,
    "date": cast_to_datetime,
    "boolean": cast_to_bool,
}


def execute_convert(
    step: ConvertStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    cast_function = CAST_FUNCTIONS[step.data_type]
    transformed_columns = {col_name: cast_function(df[col_name]) for col_name in step.columns}
    return df.assign(**transformed_columns)



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import DuplicateStep


def translate_duplicate(step: DuplicateStep) -> list[MongoStep]:
    return [
        {
            "$addFields": {(step.new_column_name): f"${step.column}"},
        }
    ]



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import SplitStep


def translate_split(step: SplitStep) -> list[MongoStep]:
    add_fields_step = {}
    for i in range(1, step.number_cols_to_keep + 1):
        add_fields_step[f"{step.column}_{i}"] = {"$arrayElemAt": ["$_vqbTmp", i - 1]}

    return [
        {"$addFields": {"_vqbTmp": {"$split": [f"${step.column}", step.delimiter]}}},
        {"$addFields": add_fields_step},
        {"$project": {"_vqbTmp": 0}},
    ]



from typing import Annotated, Any

from pydantic import BaseModel, Field

from weaverbird.pipeline.steps.hierarchy import HierarchyStep

from .steps import (
    AbsoluteValueStep,
    AbsoluteValueStepWithVariable,
    AddMissingDatesStep,
    AddMissingDatesStepWithVariables,
    AggregateStep,
    AggregateStepWithVariables,
    AppendStep,
    AppendStepWithVariable,
    ArgmaxStep,
    ArgmaxStepWithVariable,
    ArgminStep,
    ArgminStepWithVariable,
    CompareTextStep,
    CompareTextStepWithVariables,
    ConcatenateStep,
    ConcatenateStepWithVariable,
    ConvertStep,
    CumSumStep,
    CumSumStepWithVariable,
    CustomSqlStep,
    CustomStep,
    DateExtractStep,
    DateExtractStepWithVariable,
    DeleteStep,
    DissolveStep,
    DomainStep,
    DuplicateStep,
    DurationStep,
    DurationStepWithVariable,
    EvolutionStep,
    EvolutionStepWithVariable,
    FillnaStep,
    FillnaStepWithVariable,
    FilterStep,
    FilterStepWithVariables,
    FormulaStep,
    FormulaStepWithVariable,
    FromdateStep,
    IfthenelseStep,
    IfThenElseStepWithVariables,
    JoinStep,
    JoinStepWithVariable,
    LowercaseStep,
    MovingAverageStep,
    PercentageStep,
    PivotStep,
    PivotStepWithVariable,
    RankStep,
    RankStepWithVariable,
    RenameStep,
    RenameStepWithVariable,
    ReplaceStep,
    ReplaceStepWithVariable,
    ReplaceTextStep,
    ReplaceTextStepWithVariable,
    RollupStep,
    RollupStepWithVariable,
    SelectStep,
    SimplifyStep,
    SortStep,
    SplitStep,
    SplitStepWithVariable,
    StatisticsStep,
    SubstringStep,
    TextStep,
    TextStepWithVariable,
    ToDateStep,
    TopStep,
    TopStepWithVariables,
    TotalsStep,
    TotalsStepWithVariable,
    TrimStep,
    UniqueGroupsStep,
    UniqueGroupsStepWithVariable,
    UnpivotStep,
    UnpivotStepWithVariable,
    UppercaseStep,
    WaterfallStep,
    WaterfallStepWithVariable,
)

PipelineStep = Annotated[
    AbsoluteValueStep
    | AddMissingDatesStep
    | AggregateStep
    | AppendStep
    | ArgmaxStep
    | ArgminStep
    | CompareTextStep
    | ConcatenateStep
    | ConvertStep
    | CumSumStep
    | CustomSqlStep
    | CustomStep
    | DateExtractStep
    | DeleteStep
    | DissolveStep
    | DomainStep
    | DuplicateStep
    | DurationStep
    | EvolutionStep
    | FillnaStep
    | FilterStep
    | FormulaStep
    | FromdateStep
    | FromdateStep
    | HierarchyStep
    | IfthenelseStep
    | JoinStep
    | LowercaseStep
    | MovingAverageStep
    | PercentageStep
    | PivotStep
    | RankStep
    | RenameStep
    | ReplaceStep
    | ReplaceTextStep
    | RollupStep
    | SelectStep
    | SimplifyStep
    | SortStep
    | SplitStep
    | StatisticsStep
    | SubstringStep
    | TextStep
    | ToDateStep
    | TopStep
    | TotalsStep
    | TrimStep
    | UniqueGroupsStep
    | UnpivotStep
    | UppercaseStep
    | WaterfallStep,
    Field(discriminator="name"),  # noqa: F821
]


class Pipeline(BaseModel):
    steps: list[PipelineStep]

    def dict(self, *, exclude_none: bool = True, **kwargs) -> dict:
        return super().dict(exclude_none=True, **kwargs)


PipelineStepWithVariables = Annotated[
    AbsoluteValueStepWithVariable
    | AddMissingDatesStepWithVariables
    | AggregateStepWithVariables
    | AppendStepWithVariable
    | ArgmaxStepWithVariable
    | ArgminStepWithVariable
    | CompareTextStepWithVariables
    | ConcatenateStepWithVariable
    | CumSumStepWithVariable
    | DateExtractStepWithVariable
    | DurationStepWithVariable
    | EvolutionStepWithVariable
    | FillnaStepWithVariable
    | FilterStepWithVariables
    | FormulaStepWithVariable
    | IfThenElseStepWithVariables
    | JoinStepWithVariable
    | PivotStepWithVariable
    | RankStepWithVariable
    | RenameStepWithVariable
    | ReplaceStepWithVariable
    | ReplaceTextStepWithVariable
    | RollupStepWithVariable
    | SplitStepWithVariable
    | TextStepWithVariable
    | TopStepWithVariables
    | TotalsStepWithVariable
    | UniqueGroupsStepWithVariable
    | UnpivotStepWithVariable
    | WaterfallStepWithVariable,
    Field(discriminator="name"),  # noqa: F821
]


class PipelineWithVariables(BaseModel):
    steps: list[PipelineStepWithVariables | PipelineStep]

    def render(self, variables: dict[str, Any], renderer) -> Pipeline:
        # TODO it must be more efficient to render the full pipeline once
        steps_rendered = [
            step.render(variables, renderer) if hasattr(step, "render") else step  # type: ignore
            for step in self.steps
        ]
        return Pipeline(steps=steps_rendered)



from pandas import DataFrame

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import TrimStep


def execute_trim(
    step: TrimStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    for col in step.columns:
        df[col] = df[col].str.strip()
    return df



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import ReplaceStep


def translate_replace(step: ReplaceStep) -> list[MongoStep]:
    branches: list[MongoStep] = [
        {
            "case": {"$eq": [f"${step.search_column}", old_value]},
            "then": new_value,
        }
        for (old_value, new_value) in step.to_replace
    ]
    return [
        {
            "$addFields": {
                f"{step.search_column}": {
                    "$switch": {"branches": branches, "default": f"${step.search_column}"},
                }
            }
        }
    ]



import operator

import numpy as np
from pandas import DataFrame, Series

from weaverbird.pipeline.formula_ast.eval import FormulaParser
from weaverbird.pipeline.formula_ast.types import ColumnName, Expression, Operation, Operator
from weaverbird.pipeline.formula_ast.utils import unquote_string

_OP_MAP = {
    Operator.ADD: operator.add,
    Operator.SUB: operator.sub,
    Operator.MUL: operator.mul,
    Operator.DIV: operator.truediv,
    Operator.MOD: operator.mod,
}


def _eval_operation(df: DataFrame, op: Operation) -> Series:
    return _OP_MAP[op.operator](
        _eval_expression(df, op.left), _eval_expression(df, op.right)
    ).replace([np.inf, -np.inf], np.nan)


def _eval_expression(df: DataFrame, expr: Expression) -> Series:
    if isinstance(expr, Operation):
        return _eval_operation(df, expr)
    elif isinstance(expr, ColumnName):
        return df[expr.name]
    elif isinstance(expr, str):
        # Unquote the string. No need to escape quotes in it, as it will be used as is by pandas
        expr = unquote_string(expr, escape_quotes=False)
    return Series(expr, index=df.index)


def eval_formula(df: DataFrame, formula: str) -> Series:
    try:
        return _eval_expression(df, FormulaParser(formula).parse())
    except SyntaxError:  # The formula is actually a badly formatted string
        return _eval_expression(df, formula)



from typing import Literal

from pydantic import BaseModel


class Reference(BaseModel):
    type: Literal["ref"] = "ref"
    uid: str

    # Adding a prefix so is does not have the same hash as a domain whose name would be the same as
    # the uid
    def __hash__(self):
        return hash(f"__ref__{self.uid}")


PipelineOrDomainName = list[dict] | str  # can be either a domain name or a complete pipeline
PipelineOrDomainNameOrReference = PipelineOrDomainName | Reference



from typing import Literal

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps.fromdate import FromdateStep

MonthReplacement = dict[Literal["$switch"], dict[Literal["branches"], list[MongoStep]]]

_SMALL_MONTH_REPLACE = {
    "$switch": {
        "branches": [
            {"case": {"$eq": ["$_vqbTempMonth", "01"]}, "then": "Jan"},
            {"case": {"$eq": ["$_vqbTempMonth", "02"]}, "then": "Feb"},
            {"case": {"$eq": ["$_vqbTempMonth", "03"]}, "then": "Mar"},
            {"case": {"$eq": ["$_vqbTempMonth", "04"]}, "then": "Apr"},
            {"case": {"$eq": ["$_vqbTempMonth", "05"]}, "then": "May"},
            {"case": {"$eq": ["$_vqbTempMonth", "06"]}, "then": "Jun"},
            {"case": {"$eq": ["$_vqbTempMonth", "07"]}, "then": "Jul"},
            {"case": {"$eq": ["$_vqbTempMonth", "08"]}, "then": "Aug"},
            {"case": {"$eq": ["$_vqbTempMonth", "09"]}, "then": "Sep"},
            {"case": {"$eq": ["$_vqbTempMonth", "10"]}, "then": "Oct"},
            {"case": {"$eq": ["$_vqbTempMonth", "11"]}, "then": "Nov"},
            {"case": {"$eq": ["$_vqbTempMonth", "12"]}, "then": "Dec"},
        ],
        "default": None,
    },
}

_FULL_MONTH_REPLACE = {
    "$switch": {
        "branches": [
            {"case": {"$eq": ["$_vqbTempMonth", "01"]}, "then": "January"},
            {"case": {"$eq": ["$_vqbTempMonth", "02"]}, "then": "February"},
            {"case": {"$eq": ["$_vqbTempMonth", "03"]}, "then": "March"},
            {"case": {"$eq": ["$_vqbTempMonth", "04"]}, "then": "April"},
            {"case": {"$eq": ["$_vqbTempMonth", "05"]}, "then": "May"},
            {"case": {"$eq": ["$_vqbTempMonth", "06"]}, "then": "June"},
            {"case": {"$eq": ["$_vqbTempMonth", "07"]}, "then": "July"},
            {"case": {"$eq": ["$_vqbTempMonth", "08"]}, "then": "August"},
            {"case": {"$eq": ["$_vqbTempMonth", "09"]}, "then": "September"},
            {"case": {"$eq": ["$_vqbTempMonth", "10"]}, "then": "October"},
            {"case": {"$eq": ["$_vqbTempMonth", "11"]}, "then": "November"},
            {"case": {"$eq": ["$_vqbTempMonth", "12"]}, "then": "December"},
        ],
        "default": None,
    },
}


def _translate_month_year(
    step: FromdateStep, concat_separator: Literal[" ", "-"], month_dict: MongoStep
) -> list[MongoStep]:
    return [
        {
            "$addFields": {
                step.column: {
                    "$dateToString": {"date": f"${step.column}", "format": "%m-%Y"},
                },
            },
        },
        {"$addFields": {"_vqbTempArray": {"$split": [f"${step.column}", "-"]}}},
        {
            "$addFields": {
                "_vqbTempMonth": {"$arrayElemAt": ["$_vqbTempArray", 0]},
                "_vqbTempYear": {"$arrayElemAt": ["$_vqbTempArray", 1]},
            },
        },
        {
            "$addFields": {"_vqbTempMonth": month_dict},
        },
        {
            "$addFields": {
                step.column: {
                    "$concat": ["$_vqbTempMonth", concat_separator, "$_vqbTempYear"],
                },
            },
        },
        {"$project": {"_vqbTempArray": 0, "_vqbTempMonth": 0, "_vqbTempYear": 0}},
    ]


def _translate_day_month_year(
    step: FromdateStep, concat_separator: Literal[" ", "-"], month_dict: MongoStep
) -> list[MongoStep]:
    return [
        {
            "$addFields": {
                step.column: {
                    "$dateToString": {"date": f"${step.column}", "format": "%d-%m-%Y"},
                },
            },
        },
        {"$addFields": {"_vqbTempArray": {"$split": [f"${step.column}", "-"]}}},
        {
            "$addFields": {
                "_vqbTempDay": {"$arrayElemAt": ["$_vqbTempArray", 0]},
                "_vqbTempMonth": {"$arrayElemAt": ["$_vqbTempArray", 1]},
                "_vqbTempYear": {"$arrayElemAt": ["$_vqbTempArray", 2]},
            },
        },
        {
            "$addFields": {"_vqbTempMonth": month_dict},
        },
        {
            "$addFields": {
                step.column: {
                    "$concat": [
                        "$_vqbTempDay",
                        concat_separator,
                        "$_vqbTempMonth",
                        concat_separator,
                        "$_vqbTempYear",
                    ],
                },
            },
        },
        {
            "$project": {
                "_vqbTempArray": 0,
                "_vqbTempDay": 0,
                "_vqbTempMonth": 0,
                "_vqbTempYear": 0,
            }
        },
    ]


def translate_fromdate(step: FromdateStep) -> list[MongoStep]:
    if step.format == "%d %b %Y":
        return _translate_day_month_year(step, " ", _SMALL_MONTH_REPLACE)
    if step.format == "%d-%b-%Y":
        return _translate_day_month_year(step, "-", _SMALL_MONTH_REPLACE)
    if step.format == "%d %B %Y":
        return _translate_day_month_year(step, " ", _FULL_MONTH_REPLACE)
    if step.format == "%b %Y":
        return _translate_month_year(step, " ", _SMALL_MONTH_REPLACE)
    if step.format == "%b-%Y":
        return _translate_month_year(step, "-", _SMALL_MONTH_REPLACE)
    if step.format == "%B %Y":
        return _translate_month_year(step, " ", _FULL_MONTH_REPLACE)
    else:
        return [
            {
                "$addFields": {
                    step.column: {
                        "$dateToString": {"date": f"${step.column}", "format": step.format},
                    },
                },
            },
        ]



from typing import Literal, Union

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName

BASIC_DATE_PARTS = Literal[
    "year",
    "month",
    "day",
    "week",
    "quarter",
    "dayOfWeek",
    "dayOfYear",
    "isoYear",
    "isoWeek",
    "isoDayOfWeek",
    "hour",
    "minutes",
    "seconds",
    "milliseconds",
]


INTEGER_DATE_PARTS = Union[
    BASIC_DATE_PARTS,
    Literal[
        "previousYear",
        "previousMonth",
        "previousWeek",
        "previousQuarter",
        "previousIsoWeek",
    ],
]

TIMESTAMP_DATE_PARTS = Literal[
    "previousDay",
    "firstDayOfYear",
    "firstDayOfMonth",
    "firstDayOfWeek",
    "firstDayOfQuarter",
    "firstDayOfIsoWeek",
    "firstDayOfPreviousYear",
    "firstDayOfPreviousMonth",
    "firstDayOfPreviousWeek",
    "firstDayOfPreviousQuarter",
    "firstDayOfPreviousIsoWeek",
]

DATE_INFO = Union[
    INTEGER_DATE_PARTS,
    TIMESTAMP_DATE_PARTS,
]


class DateExtractStep(BaseStep):
    name: Literal["dateextract"] = "dateextract"
    column: str
    date_info: list[DATE_INFO] = Field([], alias=("dateInfo"))
    new_columns: list[ColumnName] = Field([], alias="newColumns")
    operation: BASIC_DATE_PARTS | None
    new_column_name: ColumnName | None


class DateExtractStepWithVariable(DateExtractStep, StepWithVariablesMixin):
    ...



from enum import Enum
from typing import Union

from pydantic import BaseModel

from weaverbird.pipeline.formula_ast.utils import unquote_string


class Operator(str, Enum):
    ADD = "+"
    SUB = "-"
    MUL = "*"
    DIV = "/"
    MOD = "%"


Constant = int | bool | float | str


class ColumnName(BaseModel):
    name: str
    alias: str


Expression = Union["Operation", ColumnName, Constant]


# Dataclasses do not supported recursive types for now
class Operation(BaseModel, smart_union=True):
    left: Expression
    right: Expression
    operator: Operator


Operation.update_forward_refs()


def format_expr(
    expr: Expression,
    *,
    parenthesize_operations: bool = False,
    str_quote_seq: str = "'",
    column_start_seq: str = "[",
    column_end_seq: str = "]",
    bools_as_py: bool = False,
) -> str | int | float:
    if isinstance(expr, Operation):
        left = format_expr(
            expr.left,
            parenthesize_operations=True,
            str_quote_seq=str_quote_seq,
            column_start_seq=column_start_seq,
            column_end_seq=column_end_seq,
        )
        right = format_expr(
            expr.right,
            parenthesize_operations=True,
            str_quote_seq=str_quote_seq,
            column_start_seq=column_start_seq,
            column_end_seq=column_end_seq,
        )
        str_ = f"{left} {expr.operator} {right}"
        return f"({str_})" if parenthesize_operations else str_
    elif isinstance(expr, ColumnName):
        return f"{column_start_seq}{expr.name}{column_end_seq}"
    elif isinstance(expr, str):
        # Unquoting
        return f"{str_quote_seq}{unquote_string(expr)}{str_quote_seq}"
    # bool is a subtype of int, but we don't want to have 1 and 0 replaced with 'true' or 'false'
    elif isinstance(expr, (int, float)) and not isinstance(expr, bool):
        return expr
    elif isinstance(expr, bool):
        return expr if bools_as_py else ("true" if expr else "false")



from typing import Literal

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName, TemplatedVariable

DatesGranularity = Literal["day", "week", "month", "year"]


class AddMissingDatesStep(BaseStep):
    name: Literal["addmissingdates"] = "addmissingdates"
    dates_column: ColumnName = Field(alias="datesColumn")
    dates_granularity: DatesGranularity = Field(alias="datesGranularity")
    groups: list[ColumnName] = []


class AddMissingDatesStepWithVariables(AddMissingDatesStep, StepWithVariablesMixin):
    groups: list[TemplatedVariable] | TemplatedVariable
    dates_column: TemplatedVariable = Field(alias="datesColumn")



from typing import get_args

from pandas import DataFrame, to_datetime, to_timedelta
from pandas.api.types import is_unsigned_integer_dtype

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import DateExtractStep
from weaverbird.pipeline.steps.date_extract import DATE_INFO, TIMESTAMP_DATE_PARTS

from .utils.cast import cast_to_int

OPERATIONS_MAPPING = {
    "minutes": "minute",
    "seconds": "second",
    "dayOfYear": "dayofyear",
}


def execute_date_extract(
    step: DateExtractStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    date_info: list[DATE_INFO]
    if step.operation:  # for retrocompatibility
        date_info = [step.operation]
        new_columns = [step.new_column_name or f"{step.column}_{step.operation}"]
    else:
        date_info = step.date_info
        new_columns = step.new_columns

    for dt_info, new_col in zip(date_info, new_columns):
        series_dt = to_datetime(df[step.column], utc=True)
        series_dt_accessor = series_dt.dt
        if dt_info == "week":
            # cast in float and not in int to manage NaN properly
            result = series_dt_accessor.strftime("%U").astype(float)
        elif dt_info == "dayOfWeek":
            # result should be between 1 (sunday) and 7 (saturday)
            result = (series_dt_accessor.dayofweek + 2) % 7
            result = result.replace({0: 7})
        elif dt_info == "isoYear":
            result = series_dt_accessor.isocalendar().year
        elif dt_info == "isoWeek":
            result = series_dt_accessor.isocalendar().week
        elif dt_info == "isoDayOfWeek":
            result = series_dt_accessor.isocalendar().day
        elif dt_info == "firstDayOfYear":
            result = to_datetime(DataFrame({"year": series_dt_accessor.year, "month": 1, "day": 1}))
        elif dt_info == "firstDayOfMonth":
            result = to_datetime(
                DataFrame(
                    {"year": series_dt_accessor.year, "month": series_dt_accessor.month, "day": 1}
                )
            )
        elif dt_info == "firstDayOfWeek":
            # dayofweek should be between 1 (sunday) and 7 (saturday)
            dayofweek = (series_dt_accessor.dayofweek + 2) % 7
            dayofweek = dayofweek.replace({0: 7})
            # we subtract a number of days corresponding to(dayOfWeek - 1)
            result = series_dt - to_timedelta(dayofweek - 1, unit="d")
            # the result should be returned with 0-ed time information
            result = to_datetime(result.dt.date)
        elif dt_info == "firstDayOfQuarter":
            result = to_datetime(
                DataFrame(
                    {
                        "year": series_dt_accessor.year,
                        "month": 3 * ((series_dt_accessor.month - 1) // 3) + 1,
                        "day": 1,
                    }
                )
            )
        elif dt_info == "firstDayOfIsoWeek":
            dayofweek = series_dt_accessor.isocalendar().day
            # we subtract a number of days corresponding to(dayOfWeek - 1)
            result = series_dt - to_timedelta(dayofweek - 1, unit="d")
            # the result should be returned with 0-ed time information
            result = to_datetime(result.dt.date)
        elif dt_info == "previousDay":
            result = series_dt - to_timedelta(1, unit="d")
            # the result should be returned with 0-ed time information
            result = to_datetime(result.dt.date)
        elif dt_info == "firstDayOfPreviousYear":
            result = to_datetime(
                DataFrame({"year": series_dt_accessor.year - 1, "month": 1, "day": 1})
            )
        elif dt_info == "firstDayOfPreviousMonth":
            prev_month = series_dt_accessor.month - 1
            prev_month = prev_month.replace({0: 12})
            result = to_datetime(
                DataFrame(
                    {
                        "year": series_dt_accessor.year - (prev_month == 12),
                        "month": prev_month,
                        "day": 1,
                    }
                )
            )
        elif dt_info == "firstDayOfPreviousWeek":
            prev_week_date = series_dt - to_timedelta(7, unit="d")
            # dayofweek should be between 1 (sunday) and 7 (saturday)
            dayofweek = (prev_week_date.dt.dayofweek + 2) % 7
            dayofweek = dayofweek.replace({0: 7})
            # we subtract a number of days corresponding to(dayOfWeek - 1)
            result = prev_week_date - to_timedelta(dayofweek - 1, unit="d")
            # the result should be returned with 0-ed time information
            result = to_datetime(result.dt.date)
        elif dt_info == "firstDayOfPreviousQuarter":
            first_month_of_quarter = 3 * ((series_dt_accessor.month - 1) // 3) + 1
            first_month_of_prev_q = first_month_of_quarter - 3
            first_month_of_prev_q = first_month_of_prev_q.replace({-2: 10})
            result = to_datetime(
                DataFrame(
                    {
                        "year": series_dt_accessor.year - (first_month_of_prev_q == 10),
                        "month": first_month_of_prev_q,
                        "day": 1,
                    }
                )
            )
        elif dt_info == "firstDayOfPreviousIsoWeek":
            prev_week_date = series_dt - to_timedelta(7, unit="d")
            dayofweek = prev_week_date.dt.isocalendar().day
            # we subtract a number of days corresponding to(dayOfWeek - 1)
            result = prev_week_date - to_timedelta(dayofweek - 1, unit="d")
            # the result should be returned with 0-ed time information
            result = to_datetime(result.dt.date)
        elif dt_info == "previousYear":
            result = series_dt_accessor.year - 1
        elif dt_info == "previousMonth":
            month = series_dt_accessor.month
            result = month - 1
            result = result.replace({0: 12})
        elif dt_info == "previousWeek":
            prev_week_date = series_dt - to_timedelta(7, unit="d")
            result = prev_week_date.dt.strftime("%U").astype(float)
        elif dt_info == "previousQuarter":
            result = series_dt_accessor.quarter - 1
            result = result.replace({0: 4})
        elif dt_info == "previousIsoWeek":
            prev_week_date = series_dt - to_timedelta(7, unit="d")
            result = prev_week_date.dt.isocalendar().week
        elif dt_info == "milliseconds":
            result = series_dt_accessor.microsecond / 1000
        else:
            operation = OPERATIONS_MAPPING.get(dt_info, dt_info)
            result = getattr(series_dt_accessor, operation)

        # Handle unsigned integers:
        if is_unsigned_integer_dtype(result) or dt_info not in get_args(TIMESTAMP_DATE_PARTS):
            result = cast_to_int(result)

        df = df.assign(**{new_col: result})

    return df



import logging

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.backends.mongo_translator.utils import column_to_user_variable
from weaverbird.pipeline import Pipeline, steps
from weaverbird.pipeline.steps import DomainStep, JoinStep

logger = logging.getLogger(__name__)


def translate_join(step: JoinStep) -> list[MongoStep]:
    mongo_pipeline: list[dict] = []
    right = step.right_pipeline
    right_without_domain = Pipeline(steps=[])

    if isinstance(right, str):
        right_domain = DomainStep(name="domain", domain=right)
    else:
        try:
            assert isinstance(right, list)
        except AssertionError:  # in this case right is a Reference
            raise Exception("References must be resolved before translating the pipeline")
        right_domain = DomainStep(**right[0])
        right_without_domain.steps = [
            getattr(steps, f"{s['name'].capitalize()}Step")(**s) for s in right[1:]
        ]

    mongo_let: dict[str, str] = {}
    mongo_expr_and: list[dict[str, list[str]]] = []

    for left_on, right_on in step.on:
        mongo_let[column_to_user_variable(left_on)] = f"${left_on}"
        mongo_expr_and.append({"$eq": [f"${right_on}", f"$${column_to_user_variable(left_on)}"]})

    from weaverbird.backends.mongo_translator.mongo_pipeline_translator import translate_pipeline

    mongo_pipeline.append(
        {
            "$lookup": {
                "from": right_domain.domain,
                "let": mongo_let,
                "pipeline": translate_pipeline(right_without_domain)
                + [{"$match": {"$expr": {"$and": mongo_expr_and}}}],
                "as": "_vqbJoinKey",
            }
        }
    )
    if step.type == "inner":
        mongo_pipeline.append({"$unwind": "$_vqbJoinKey"})
    elif step.type == "left":
        mongo_pipeline.append(
            {"$unwind": {"path": "$_vqbJoinKey", "preserveNullAndEmptyArrays": True}}
        )
    else:
        mongo_pipeline.append({"$match": {"_vqbJoinKey": {"$eq": []}}})
        mongo_pipeline.append(
            {"$unwind": {"path": "$_vqbJoinKey", "preserveNullAndEmptyArrays": True}}
        )

    mongo_pipeline.append(
        {
            "$replaceRoot": {"newRoot": {"$mergeObjects": ["$_vqbJoinKey", "$$ROOT"]}},
        }
    )
    mongo_pipeline.append(
        {
            "$project": {"_vqbJoinKey": 0},
        },
    )
    return mongo_pipeline



from typing import TYPE_CHECKING, Any, TypeVar

from pypika import Field, Query, Table, functions
from pypika.enums import Dialects
from pypika.queries import QueryBuilder
from pypika.terms import Case, CustomFunction, Function, Interval, LiteralValue, Term

from weaverbird.backends.pypika_translator.dialects import SQLDialect
from weaverbird.backends.pypika_translator.operators import FromDateOp, RegexOp
from weaverbird.backends.pypika_translator.translators.base import (
    DataTypeMapping,
    DateFormatMapping,
    SQLTranslator,
    StepContext,
)
from weaverbird.pipeline.steps.date_extract import DATE_INFO

Self = TypeVar("Self", bound="GoogleBigQueryTranslator")

if TYPE_CHECKING:
    from weaverbird.pipeline.steps import SplitStep, ToDateStep


class GBQDateTrunc(Function):
    """The date format and field parameters are inverted in GBQ"""

    def __init__(self, date_format: str, field: Field, alias: str | None = None):
        super().__init__("DATE_TRUNC", field, LiteralValue(date_format.upper()), alias=alias)


class GBQParseDateTime(Function):
    def __init__(self, term: str | Field, date_format: str) -> None:
        # Inverting date_format and term, because GBQ takes format first
        super().__init__("PARSE_DATETIME", date_format, term)


class GBQSplit(Function):
    def __init__(self, field: Field, delimiter: str | None = None) -> None:
        # Empty string is the same as None here
        args = ("SPLIT", field, Term.wrap_constant(delimiter)) if delimiter else ("SPLIT", field)
        super().__init__(*args)


class GoogleBigQueryQuery(Query):
    @classmethod
    def _builder(cls, **kwargs: Any) -> "GoogleBigQueryQueryBuilder":
        return GoogleBigQueryQueryBuilder(**kwargs)


class GoogleBigQueryQueryBuilder(QueryBuilder):
    QUOTE_CHAR = "`"
    SECONDARY_QUOTE_CHAR = "'"
    ALIAS_QUOTE_CHAR = None
    QUERY_ALIAS_QUOTE_CHAR = None
    QUERY_CLS = GoogleBigQueryQuery

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(dialect=SQLDialect.GOOGLEBIGQUERY, **kwargs)


class GoogleBigQueryDateAdd(Function):
    def __init__(self, *, target_column: Field, interval: Interval) -> None:
        super().__init__("DATE_ADD", target_column, interval)


class GoogleBigQueryTranslator(SQLTranslator):
    DIALECT = SQLDialect.GOOGLEBIGQUERY
    QUERY_CLS = GoogleBigQueryQuery
    DATA_TYPE_MAPPING = DataTypeMapping(
        boolean="BOOLEAN",
        date="DATE",
        float="FLOAT64",
        integer="INTEGER",
        text="STRING",
        datetime="DATETIME",
        timestamp="TIMESTAMP",
    )
    DATE_FORMAT_MAPPING = DateFormatMapping(
        # https://cloud.google.com/bigquery/docs/reference/standard-sql/format-elements#format_elements_date_time
        day_number="%d",
        month_number="%m",
        month_short="%b",
        month_full="%B",
        year="%Y",
    )
    SUPPORT_ROW_NUMBER = True
    SUPPORT_SPLIT_PART = False
    SUPPORT_UNPIVOT = True
    FROM_DATE_OP = FromDateOp.TO_CHAR
    REGEXP_OP = RegexOp.REGEXP_CONTAINS

    @classmethod
    def _add_date(
        cls, *, target_column: Field, duration: int, unit: str, dialect: Dialects | None = None
    ) -> Term:
        return GoogleBigQueryDateAdd(
            target_column=target_column,
            # Cheating a bit here: MySQL's syntax is compatible with GBQ for intervals
            interval=Interval(**{unit: duration, "dialect": Dialects.MYSQL}),
        )

    @classmethod
    def _day_of_week(cls, target_column: Field) -> Term:
        # GBQ takes 'dayofweek' rather than 'dow' and returns 1 for sunday rather than 0
        return functions.Extract("dayofweek", target_column) - 1

    def split(
        self: Self,
        *,
        builder: "QueryBuilder",
        prev_step_name: str,
        columns: list[str],
        step: "SplitStep",
    ) -> StepContext:
        col_field: Field = Table(prev_step_name)[step.column]

        safe_offset = CustomFunction("SAFE_OFFSET", ["index"])

        # Sub-optimal, could do that in two sub_queries, one for splitting to a temp array col, and
        # another one to select eveything needed from the array col rather than splitting N times
        def gen_splitted_cols():
            for i in range(step.number_cols_to_keep):
                split_str = GBQSplit(col_field, step.delimiter).get_sql(
                    quote_char=GoogleBigQueryQueryBuilder.QUOTE_CHAR
                )
                safe_offset_str = safe_offset(i).get_sql(
                    quote_char=GoogleBigQueryQueryBuilder.QUOTE_CHAR
                )
                # LiteralValue is ugly, but it does not seem like pypika supports "[]" array
                # accessing, and GBQ does not seem to provide functions to access array value.
                #
                # The IfNull is required because other backends use SPLIT_PART, which will return an
                # empty string rather than NULL
                yield functions.IfNull(LiteralValue(f"{split_str}[{safe_offset_str}]"), "").as_(
                    f"{step.column}_{i + 1}"
                )

        splitted_cols = list(gen_splitted_cols())
        query: "QueryBuilder" = self.QUERY_CLS.from_(prev_step_name).select(
            *columns, *splitted_cols
        )
        return StepContext(query, columns + splitted_cols)

    @classmethod
    def _date_trunc(cls, date_part: str, target_column: Field) -> Term:
        return GBQDateTrunc(date_part, target_column)

    @classmethod
    def _get_date_extract_func(cls, *, date_unit: DATE_INFO, target_column: Field) -> Term:
        if date_unit == "dayOfYear":
            # GBQ takes 'dayofyear' rather than 'doy'
            return functions.Extract("dayofyear", target_column)
        if date_unit == "week":
            return functions.Extract("isoweek", target_column)
        if date_unit == "previousWeek":
            return functions.Extract(
                "isoweek", cls._add_date(target_column=target_column, unit="weeks", duration=-1)
            )

        if date_unit == "isoWeek":
            return (
                Case()
                .when(functions.Extract("WEEK(MONDAY)", target_column) == 0, 52)
                .else_(functions.Extract("WEEK(MONDAY)", target_column))
            )
        if date_unit == "previousIsoWeek":
            return (
                Case()
                .when(
                    functions.Extract(
                        "WEEK(MONDAY)",
                        cls._add_date(target_column=target_column, unit="weeks", duration=-1),
                    )
                    == 0,
                    52,
                )
                .else_(
                    functions.Extract(
                        "WEEK(MONDAY)",
                        cls._add_date(target_column=target_column, unit="weeks", duration=-1),
                    )
                )
            )

        if date_unit == "isoYear":
            return (
                Case()
                .when(
                    functions.Extract("WEEK(MONDAY)", target_column) != 0,
                    functions.Extract("year", target_column),
                )
                .else_(functions.Extract("year", target_column) - 1)
            )
        if date_unit == "firstDayOfWeek":
            return cls._date_trunc("WEEK", target_column)
        if date_unit == "firstDayOfPreviousWeek":
            return cls._add_date(
                target_column=cls._date_trunc("WEEK", target_column), duration=-1, unit="weeks"
            )
        if date_unit == "firstDayOfIsoWeek":
            return cls._date_trunc("ISOWEEK", target_column)
        if date_unit == "firstDayOfPreviousIsoWeek":
            return cls._add_date(
                target_column=cls._date_trunc("ISOWEEK", target_column), duration=-1, unit="weeks"
            )
        return super()._get_date_extract_func(date_unit=date_unit, target_column=target_column)

    def todate(
        self: Self,
        *,
        builder: "QueryBuilder",
        prev_step_name: str,
        columns: list[str],
        step: "ToDateStep",
    ) -> StepContext:
        col_field = Table(prev_step_name)[step.column]
        if step.format is not None:
            date_selection = GBQParseDateTime(col_field, step.format)
        else:
            date_selection = functions.Cast(col_field, "DATETIME")

        query: "QueryBuilder" = self.QUERY_CLS.from_(prev_step_name).select(
            *(c for c in columns if c != step.column),
            date_selection.as_(step.column),
        )
        return StepContext(query, columns)


SQLTranslator.register(GoogleBigQueryTranslator)


class ParseDatetime(functions.Function):
    def __init__(self, format: str, term: str | Field, alias: str | None = None) -> None:
        super().__init__("parse_datetime", format, term, alias=alias)



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import RenameStep


def translate_rename(step: RenameStep) -> list[MongoStep]:
    return [
        {"$addFields": {to_rename[1]: f"${to_rename[0]}" for to_rename in step.to_rename}},
        {
            "$project": {
                to_rename[0]: 0
                for to_rename in step.to_rename
                if to_rename[0] not in [to_rename[1] for to_rename in step.to_rename]
            }
        },
    ]



from typing import Literal

from pydantic import Field, validator

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.types import ColumnName


class PercentageStep(BaseStep):
    name: Literal["percentage"] = "percentage"
    column: ColumnName
    group: list[ColumnName] = Field(default=[])
    new_column_name: ColumnName | None = Field(alias="newColumnName", default=None)

    @validator("new_column_name", always=True)
    def set_new_column_name_if_not_set(cls, v, values):
        if v is None:
            return f"{values['column']}_PCT"
        return v



from typing import Literal

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import TemplatedVariable


class PivotStep(BaseStep):
    name: Literal["pivot"] = "pivot"
    index: list[str]
    column_to_pivot: str
    value_column: str
    agg_function: Literal["sum", "avg", "count", "min", "max"]


class PivotStepWithVariable(PivotStep, StepWithVariablesMixin):
    index: TemplatedVariable | list[TemplatedVariable]



from typing import Any, Literal

from pydantic import Field, root_validator

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName


class FillnaStep(BaseStep):
    name: Literal["fillna"] = "fillna"
    columns: list[ColumnName] = Field(min_items=1)
    value: Any

    @root_validator(pre=True)
    def handle_legacy_syntax(cls, values):
        if "column" in values:
            values["columns"] = [values.pop("column")]
        return values


class FillnaStepWithVariable(FillnaStep, StepWithVariablesMixin):
    ...



from typing import Any

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import AggregateStep

_ID_COLUMN = "_id"
_NEW_ID_COLUMN = "__id"


def translate_aggregate(step: AggregateStep) -> list[MongoStep]:
    idblock = {col: f"${col}" for col in step.on}
    group: dict[str, dict] = {}
    project: dict[str, Any] = {}
    add_fields = {}
    group[_ID_COLUMN] = idblock

    for agg_f_step in step.aggregations:
        cols = agg_f_step.columns
        # _id is a name reserved by mongo. If for some reason, a user wants to aggregate the _id
        # column, the aggregation result will be stored in a new __id column
        new_cols = [col if col != _ID_COLUMN else _NEW_ID_COLUMN for col in agg_f_step.new_columns]

        # There is no `$count` operator in Mongo, we have to `$sum` 1s to get
        # an equivalent result
        if agg_f_step.agg_function == "count":
            for i in range(len(cols)):
                group[new_cols[i]] = {"$sum": 1}
        elif agg_f_step.agg_function == "count distinct":
            for i in range(len(cols)):
                group[new_cols[i]] = {"$addToSet": f"${cols[i]}"}
                add_fields[new_cols[i]] = {"$size": f"${new_cols[i]}"}
        else:
            for i in range(len(cols)):
                group[new_cols[i]] = {f"${agg_f_step.agg_function}": f"${cols[i]}"}

    add_fields_to_add_to_pipeline = [{"$addFields": add_fields}] if len(add_fields) else []
    if step.keep_original_granularity:
        group["_vqbDocsArray"] = {"$push": "$$ROOT"}
        return [
            {"$group": group},
            *add_fields_to_add_to_pipeline,
            {"$unwind": "$_vqbDocsArray"},
            {"$replaceRoot": {"newRoot": {"$mergeObjects": ["$_vqbDocsArray", "$$ROOT"]}}},
            {"$project": {"_vqbDocsArray": 0}},
            {"$sort": {_ID_COLUMN: 1}},
        ]
    else:
        for group_key in group.keys():
            if group_key == _ID_COLUMN:
                for id_key in group[group_key]:
                    project[id_key] = f"${_ID_COLUMN}.{id_key}"
            else:
                project[group_key] = 1
        return [
            {"$group": group},
            *add_fields_to_add_to_pipeline,
            {"$project": project},
            {"$sort": {_ID_COLUMN: 1}},
        ]



from typing import Literal

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep, to_camelcase
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName, TemplatedVariable


class ReplaceTextStep(BaseStep):
    class Config:
        alias_generator = to_camelcase

    name: Literal["replacetext"] = "replacetext"
    search_column: ColumnName
    old_str: str
    new_str: str = Field(default="")


class ReplaceTextStepWithVariable(ReplaceTextStep, StepWithVariablesMixin):
    old_str: TemplatedVariable



from typing import Literal

from pydantic import BaseModel

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.types import ColumnName


class ColumnSort(BaseModel):
    column: ColumnName
    order: Literal["asc", "desc"]


class SortStep(BaseStep):
    name: Literal["sort"] = "sort"
    columns: list[ColumnSort]



from typing import Literal

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.types import ColumnName


class LowercaseStep(BaseStep):
    name: Literal["lowercase"] = "lowercase"
    column: ColumnName



from typing import Literal

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin

from .utils.combination import PipelineOrDomainNameOrReference


class AppendStep(BaseStep):
    name: Literal["append"] = "append"
    pipelines: list[PipelineOrDomainNameOrReference]


class AppendStepWithVariable(AppendStep, StepWithVariablesMixin):
    ...



from typing import Literal

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName


class FormulaStep(BaseStep):
    name: Literal["formula"] = "formula"
    new_column: ColumnName
    formula: str


class FormulaStepWithVariable(FormulaStep, StepWithVariablesMixin):
    ...



from typing import Literal

from pydantic import BaseModel, Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.types import ColumnName

DUMB_GROUPBY_COLUMN_NAME = "__dumb_groupby_column_name__"

Statistics = Literal["count", "max", "min", "average", "variance", "standard deviation"]


class Quantile(BaseModel):
    label: str | None
    nth: int
    order: int


class StatisticsStep(BaseStep):
    name: Literal["statistics"] = "statistics"
    column: ColumnName
    groupby_columns: list[ColumnName] = Field([], alias="groupbyColumns")
    statistics: list[Statistics]
    # Array of quantiles. Examples:
    # - median is 1rst quantile of order 2
    # - last decile is 9th quantile of order 10
    quantiles: list[Quantile]



from typing import Literal

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName, TemplatedVariable


class CompareTextStep(BaseStep):
    name: Literal["comparetext"] = "comparetext"
    new_column_name: ColumnName = Field(alias="newColumnName")
    str_col_1: ColumnName = Field(alias="strCol1")
    str_col_2: ColumnName = Field(alias="strCol2")


class CompareTextStepWithVariables(CompareTextStep, StepWithVariablesMixin):
    str_col_1: TemplatedVariable = Field(alias="strCol1")
    str_col_2: TemplatedVariable = Field(alias="strCol2")



from typing import Literal

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName, TemplatedVariable


class UniqueGroupsStep(BaseStep):
    name: Literal["uniquegroups"] = "uniquegroups"
    on: list[ColumnName]


class UniqueGroupsStepWithVariable(UniqueGroupsStep, StepWithVariablesMixin):
    on: TemplatedVariable | list[TemplatedVariable]



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import ConcatenateStep


def _convert_to_type(input: str | dict, type_: str) -> dict[str, str | dict]:
    return {"$convert": {"input": input, "to": type_}}


def translate_concatenate(step: ConcatenateStep) -> list[MongoStep]:
    # NOTE: Translated from "../src/lib/translators/mongo.ts"'s
    # concatenate() step. The difference is that we're using mongo's
    # $convert operator here, which requires mongo>=4.0
    concat_block: list[dict | str] = [_convert_to_type(f"${step.columns[0]}", "string")]
    for col in step.columns[1:]:
        concat_block += [step.separator, _convert_to_type(f"${col}", "string")]

    return [{"$addFields": {step.new_column_name: {"$concat": concat_block}}}]



from typing import Literal

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName

from .utils.combination import PipelineOrDomainNameOrReference

JoinColumnsPair = tuple[ColumnName, ColumnName]


class JoinStep(BaseStep):
    name: Literal["join"] = "join"
    right_pipeline: PipelineOrDomainNameOrReference
    type: Literal["left", "inner", "left outer"]
    on: list[JoinColumnsPair] = Field(..., min_items=1)


class JoinStepWithVariable(JoinStep, StepWithVariablesMixin):
    ...



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import UnpivotStep


def translate_unpivot(step: UnpivotStep) -> list[MongoStep]:
    # project_cols to be included in Mongo $project steps
    project_cols = {col: f"${col}" for col in step.keep}
    # object_to_array to be included in the first Mongo $project step
    object_to_array = {col: {"$ifNull": [f"${col}", None]} for col in step.unpivot}

    mongo_pipeline: list[dict] = [
        {
            "$project": {**project_cols, "_vqbToUnpivot": {"$objectToArray": object_to_array}},
        },
        {
            "$unwind": "$_vqbToUnpivot",
        },
        {
            "$project": {
                **project_cols,
                step.unpivot_column_name: "$_vqbToUnpivot.k",
                step.value_column_name: "$_vqbToUnpivot.v",
            },
        },
    ]

    if step.dropna:
        mongo_pipeline.append({"$match": {step.value_column_name: {"$ne": None}}})

    return mongo_pipeline



from typing import Literal

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.types import ColumnName


class SubstringStep(BaseStep):
    name: Literal["substring"] = "substring"
    column: ColumnName
    new_column_name: ColumnName | None = Field(alias="newColumnName")
    start_index: int
    end_index: int



from typing import Literal

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.types import ColumnName


class UppercaseStep(BaseStep):
    name: Literal["uppercase"] = "uppercase"
    column: ColumnName



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import SelectStep


def translate_select(step: SelectStep) -> list[MongoStep]:
    return [{"$project": {f"{column}": 1 for column in step.columns}}]



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import DateExtractStep


def _truncate_date_to_day(expr: dict | str) -> MongoStep:
    return {
        "$dateTrunc": {
            "unit": "day",
            "date": expr,
        },
    }


def _extract_quarter(step: DateExtractStep) -> MongoStep:
    return {
        "$switch": {
            "branches": [
                {"case": {"$lte": [{"$divide": [{"$month": f"${step.column}"}, 3]}, 1]}, "then": 1},
                {"case": {"$lte": [{"$divide": [{"$month": f"${step.column}"}, 3]}, 2]}, "then": 2},
                {"case": {"$lte": [{"$divide": [{"$month": f"${step.column}"}, 3]}, 3]}, "then": 3},
            ],
            "default": 4,
        },
    }


def _extract_first_day_of_year(step: DateExtractStep) -> MongoStep:
    return {
        "$dateFromParts": {"year": {"$year": f"${step.column}"}, "month": 1, "day": 1},
    }


def _extract_first_day_of_month(step: DateExtractStep) -> MongoStep:
    return {
        "$dateFromParts": {
            "year": {"$year": f"${step.column}"},
            "month": {"$month": f"${step.column}"},
            "day": 1,
        },
    }


def _extract_first_day_of_week(step: DateExtractStep) -> MongoStep:
    return _truncate_date_to_day(
        {
            # We subtract to the target date a number of days corresponding to (dayOfWeek - 1)
            "$subtract": [
                f"${step.column}",
                {
                    "$multiply": [
                        {"$subtract": [{"$dayOfWeek": f"${step.column}"}, 1]},
                        24 * 60 * 60 * 1000,
                    ],
                },
            ],
        }
    )


def _extract_first_day_of_quarter(step: DateExtractStep) -> MongoStep:
    return {
        "$dateFromParts": {
            "year": {"$year": f"${step.column}"},
            "month": {
                "$switch": {
                    "branches": [
                        {
                            "case": {"$lte": [{"$divide": [{"$month": f"${step.column}"}, 3]}, 1]},
                            "then": 1,
                        },
                        {
                            "case": {"$lte": [{"$divide": [{"$month": f"${step.column}"}, 3]}, 2]},
                            "then": 4,
                        },
                        {
                            "case": {"$lte": [{"$divide": [{"$month": f"${step.column}"}, 3]}, 3]},
                            "then": 7,
                        },
                    ],
                    "default": 10,
                },
            },
            "day": 1,
        },
    }


def _extract_first_day_of_iso_week(step: DateExtractStep) -> MongoStep:
    return _truncate_date_to_day(
        {
            # We subtract to the target date a number of days corresponding to (isoDayOfWeek - 1)
            "$subtract": [
                f"${step.column}",
                {
                    "$multiply": [
                        {"$subtract": [{"$isoDayOfWeek": f"${step.column}"}, 1]},
                        24 * 60 * 60 * 1000,
                    ],
                },
            ],
        }
    )


def _extract_previous_day(step: DateExtractStep) -> MongoStep:
    return _truncate_date_to_day(
        {
            # We subtract to the target date 1 day in milliseconds
            "$subtract": [f"${step.column}", 24 * 60 * 60 * 1000],
        }
    )


def _extract_first_day_of_previous_year(step: DateExtractStep) -> MongoStep:
    return {
        "$dateFromParts": {
            "year": {"$subtract": [{"$year": f"${step.column}"}, 1]},
            "month": 1,
            "day": 1,
        },
    }


def _extract_first_day_of_previous_month(step: DateExtractStep) -> MongoStep:
    return {
        "$dateFromParts": {
            "year": {
                "$cond": [
                    {"$eq": [{"$month": f"${step.column}"}, 1]},
                    {"$subtract": [{"$year": f"${step.column}"}, 1]},
                    {"$year": f"${step.column}"},
                ],
            },
            "month": {
                "$cond": [
                    {"$eq": [{"$month": f"${step.column}"}, 1]},
                    12,
                    {"$subtract": [{"$month": f"${step.column}"}, 1]},
                ],
            },
            "day": 1,
        },
    }


def _extract_first_day_of_previous_week(step: DateExtractStep) -> MongoStep:
    return _truncate_date_to_day(
        {
            # We subtract to the target date a number of days corresponding to (dayOfWeek - 1)
            "$subtract": [
                {"$subtract": [f"${step.column}", 7 * 24 * 60 * 60 * 1000]},
                {
                    "$multiply": [
                        {"$subtract": [{"$dayOfWeek": f"${step.column}"}, 1]},
                        24 * 60 * 60 * 1000,
                    ],
                },
            ],
        }
    )


def _extract_first_day_of_previous_quarter(step: DateExtractStep) -> MongoStep:
    return {
        "$dateFromParts": {
            "year": {
                "$cond": [
                    {"$lte": [{"$divide": [{"$month": f"${step.column}"}, 3]}, 1]},
                    {"$subtract": [{"$year": f"${step.column}"}, 1]},
                    {"$year": f"${step.column}"},
                ],
            },
            "month": {
                "$switch": {
                    "branches": [
                        {
                            "case": {"$lte": [{"$divide": [{"$month": f"${step.column}"}, 3]}, 1]},
                            "then": 10,
                        },
                        {
                            "case": {"$lte": [{"$divide": [{"$month": f"${step.column}"}, 3]}, 2]},
                            "then": 1,
                        },
                        {
                            "case": {"$lte": [{"$divide": [{"$month": f"${step.column}"}, 3]}, 3]},
                            "then": 4,
                        },
                    ],
                    "default": 7,
                },
            },
            "day": 1,
        },
    }


def _extract_first_day_of_previous_iso_week(step: DateExtractStep) -> MongoStep:
    return _truncate_date_to_day(
        {
            # We subtract to the target date a number of days corresponding to (isoDayOfWeek - 1)
            "$subtract": [
                {"$subtract": [f"${step.column}", 7 * 24 * 60 * 60 * 1000]},
                {
                    "$multiply": [
                        {"$subtract": [{"$isoDayOfWeek": f"${step.column}"}, 1]},
                        24 * 60 * 60 * 1000,
                    ],
                },
            ],
        }
    )


def _extract_previous_year(step: DateExtractStep) -> MongoStep:
    return {
        "$subtract": [{"$year": f"${step.column}"}, 1],
    }


def _extract_previous_month(step: DateExtractStep) -> MongoStep:
    return {
        "$cond": [
            {"$eq": [{"$month": f"${step.column}"}, 1]},
            12,
            {"$subtract": [{"$month": f"${step.column}"}, 1]},
        ],
    }


def _extract_previous_week(step: DateExtractStep) -> MongoStep:
    return {
        # We subtract to the target date 7 days in milliseconds
        "$week": {"$subtract": [f"${step.column}", 7 * 24 * 60 * 60 * 1000]},
    }


def _extract_previous_quarter(step: DateExtractStep) -> MongoStep:
    return {
        "$switch": {
            "branches": [
                {"case": {"$lte": [{"$divide": [{"$month": f"${step.column}"}, 3]}, 1]}, "then": 4},
                {"case": {"$lte": [{"$divide": [{"$month": f"${step.column}"}, 3]}, 2]}, "then": 1},
                {"case": {"$lte": [{"$divide": [{"$month": f"${step.column}"}, 3]}, 3]}, "then": 2},
            ],
            "default": 3,
        },
    }


def _extract_previous_iso_week(step: DateExtractStep) -> MongoStep:
    return {
        # We subtract to the target date 7 days in milliseconds
        "$isoWeek": {"$subtract": [f"${step.column}", 7 * 24 * 60 * 60 * 1000]},
    }


_ADVANCED_DATE_EXTRACT_MAP = {
    "quarter": _extract_quarter,
    "firstDayOfYear": _extract_first_day_of_year,
    "firstDayOfMonth": _extract_first_day_of_month,
    "firstDayOfWeek": _extract_first_day_of_week,
    "firstDayOfQuarter": _extract_first_day_of_quarter,
    "firstDayOfIsoWeek": _extract_first_day_of_iso_week,
    "previousDay": _extract_previous_day,
    "firstDayOfPreviousYear": _extract_first_day_of_previous_year,
    "firstDayOfPreviousMonth": _extract_first_day_of_previous_month,
    "firstDayOfPreviousWeek": _extract_first_day_of_previous_week,
    "firstDayOfPreviousQuarter": _extract_first_day_of_previous_quarter,
    "firstDayOfPreviousIsoWeek": _extract_first_day_of_previous_iso_week,
    "previousYear": _extract_previous_year,
    "previousMonth": _extract_previous_month,
    "previousWeek": _extract_previous_week,
    "previousQuarter": _extract_previous_quarter,
    "previousIsoWeek": _extract_previous_iso_week,
}

_DATE_EXTRACT_MAP = {
    "year": "$year",
    "month": "$month",
    "day": "$dayOfMonth",
    "week": "$week",
    "dayOfYear": "$dayOfYear",
    "dayOfWeek": "$dayOfWeek",
    "isoYear": "$isoWeekYear",
    "isoWeek": "$isoWeek",
    "isoDayOfWeek": "$isoDayOfWeek",
    "hour": "$hour",
    "minutes": "$minute",
    "seconds": "$second",
    "milliseconds": "$millisecond",
}


def translate_date_extract(step: DateExtractStep) -> list[MongoStep]:
    new_columns = []
    add_fields = {}

    date_info: list
    # For retrocompatibility
    if step.operation:
        date_info = [step.operation] if step.operation else step.date_info
        new_columns = [
            step.new_column_name if step.new_column_name else f"{step.column}_{step.operation}"
        ]
    else:
        date_info = step.date_info.copy()
        new_columns = step.new_columns.copy()

    for i, d in enumerate(date_info):
        if d in _ADVANCED_DATE_EXTRACT_MAP:
            add_fields[new_columns[i]] = _ADVANCED_DATE_EXTRACT_MAP[d](step)
        else:
            add_fields[new_columns[i]] = {
                _DATE_EXTRACT_MAP[d]: f"${step.column}",
            }

    return [{"$addFields": add_fields}]



from typing import TYPE_CHECKING

from weaverbird.backends.pypika_translator.dialects import SQLDialect

if TYPE_CHECKING:
    from .base import SQLTranslator

ALL_TRANSLATORS: dict[SQLDialect, type["SQLTranslator"]] = {}

from .athena import AthenaTranslator  # noqa
from .googlebigquery import GoogleBigQueryTranslator  # noqa
from .mysql import MySQLTranslator  # noqa
from .postgresql import PostgreSQLTranslator  # noqa
from .redshift import RedshiftTranslator  # noqa
from .snowflake import SnowflakeTranslator  # noqa



from typing import Any, Literal

from pandas import DataFrame, concat

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import AggregateStep

AggregateFn = Literal[
    "avg",
    "sum",
    "min",
    "max",
    "count",
    "count distinct",
    "first",
    "last",
    "count distinct including empty",
]

functions_aliases = {
    "avg": "mean",
    "count distinct": "nunique",
    "count distinct including empty": len,
}


def get_aggregate_fn(agg_function: str) -> Any:
    if agg_function in functions_aliases:
        return functions_aliases[agg_function]
    return agg_function


def execute_aggregate(
    step: AggregateStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    group_by_columns = step.on

    # if no group is specified, we create a pseudo column with a single value
    if len(group_by_columns) == 0:
        group_by_columns = ["__VQB__GROUP_BY__"]
        df = df.assign(**{group_by_columns[0]: True})

    grouped_by_df = df.groupby(group_by_columns, dropna=False)
    aggregated_cols = []

    if len(step.aggregations) == 0:
        df_result = grouped_by_df.first().reset_index()[group_by_columns]

    else:
        for aggregation in step.aggregations:
            for col, new_col in zip(aggregation.columns, aggregation.new_columns):
                agg_serie = (
                    grouped_by_df[col]
                    .agg(get_aggregate_fn(aggregation.agg_function))
                    .rename(new_col)
                )
                aggregated_cols.append(agg_serie)

        df_result = concat(aggregated_cols, axis=1).reset_index()

    # it is faster this way, than to transform the original df
    if step.keep_original_granularity:
        df_result = df.merge(df_result, on=group_by_columns, how="left")

    # we do not want the pseudo column to ever leave this function
    if len(step.on) == 0:
        del df_result[group_by_columns[0]]
    return df_result



from typing import Literal

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName, TemplatedVariable


class RankStep(BaseStep):
    name: Literal["rank"] = "rank"
    value_col: ColumnName = Field(alias="valueCol")
    order: Literal["asc", "desc"]
    method: Literal["standard", "dense"]
    groupby: list[ColumnName] = []
    new_column_name: ColumnName | None = Field(None, alias="newColumnName")


class RankStepWithVariable(RankStep, StepWithVariablesMixin):
    groupby: TemplatedVariable | list[TemplatedVariable]



from typing import Any, TypeVar

from pypika.dialects import QueryBuilder, SnowflakeQueryBuilder
from pypika.enums import Dialects
from pypika.queries import Query, Selectable
from pypika.terms import Field, Term
from pypika.utils import builder, format_quotes

from weaverbird.backends.pypika_translator.dialects import SQLDialect
from weaverbird.backends.pypika_translator.operators import FromDateOp, RegexOp, ToDateOp
from weaverbird.backends.pypika_translator.translators.base import (
    DataTypeMapping,
    DateAddWithoutUnderscore,
    DateFormatMapping,
    SQLTranslator,
)

Self = TypeVar("Self", bound="SQLTranslator")


class QuotedSnowflakeQueryBuilder(SnowflakeQueryBuilder):
    QUOTE_CHAR = '"'

    @builder
    def with_(self, selectable: Selectable, name: str) -> "QueryBuilder":
        return super().with_(selectable, format_quotes(name, self.QUOTE_CHAR))


class SnowflakeQuery(Query):
    @classmethod
    def _builder(cls, **kwargs: Any) -> "QuotedSnowflakeQueryBuilder":
        return QuotedSnowflakeQueryBuilder(**kwargs)


class SnowflakeTranslator(SQLTranslator):
    DIALECT = SQLDialect.SNOWFLAKE
    QUERY_CLS = SnowflakeQuery
    DATA_TYPE_MAPPING = DataTypeMapping(
        boolean="BOOLEAN",
        date="DATE",
        float="DOUBLE PRECISION",
        integer="INTEGER",
        text="TEXT",
        datetime="TIMESTAMP",
        timestamp="TIMESTAMP",
    )
    DATE_FORMAT_MAPPING = DateFormatMapping(
        # https://docs.snowflake.com/en/sql-reference/data-types-datetime#date-and-time-formats
        day_number="DD",
        month_number="MM",
        month_short="MON",
        month_full="MON",
        year="YYYY",
    )
    SUPPORT_ROW_NUMBER = True
    SUPPORT_SPLIT_PART = True
    SUPPORT_UNPIVOT = True
    FROM_DATE_OP = FromDateOp.TO_CHAR
    REGEXP_OP = RegexOp.REGEXP
    TO_DATE_OP = ToDateOp.TO_TIMESTAMP_NTZ

    @classmethod
    def _add_date(
        cls, *, target_column: Field, duration: int, unit: str, dialect: Dialects | None = None
    ) -> Term:
        return DateAddWithoutUnderscore(
            date_part=unit.removesuffix("s"), interval=duration, term=target_column
        )


SQLTranslator.register(SnowflakeTranslator)



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import TrimStep


def translate_trim(step: TrimStep) -> list[MongoStep]:
    add_fields = {x: {"$trim": {"input": f"${x}"}} for x in step.columns}

    return [{"$addFields": add_fields}]



from typing import Literal

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.types import ColumnName


class ConvertStep(BaseStep):
    name: Literal["convert"] = "convert"
    columns: list[ColumnName]
    data_type: Literal["integer", "float", "text", "date", "boolean"]



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import SortStep


def translate_sort(step: SortStep) -> list[MongoStep]:
    return [
        {
            "$sort": {
                sort_column.column: 1 if sort_column.order == "asc" else -1
                for sort_column in step.columns
            }
        }
    ]



from typing import Literal

from pydantic import Field, root_validator

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName, TemplatedVariable


class CumSumStep(BaseStep):
    name: Literal["cumsum"] = "cumsum"
    to_cumsum: list[tuple[str, str | None]] = Field(..., alias="toCumSum")
    reference_column: ColumnName = Field(..., alias="referenceColumn")
    groupby: list[ColumnName] | None

    @root_validator(pre=True)
    def handle_legacy_syntax(cls, values):
        if "valueColumn" in values:
            values["value_column"] = values.pop("valueColumn")
        if "newColumn" in values:
            values["new_column"] = values.pop("newColumn")

        if "value_column" in values:
            values["to_cumsum"] = [(values.pop("value_column"), values.pop("new_column", None))]
        return values


class CumSumStepWithVariable(CumSumStep, StepWithVariablesMixin):
    to_cumsum: TemplatedVariable | list[tuple[TemplatedVariable, TemplatedVariable]] = Field(
        ..., alias="toCumSum"
    )



from typing import Literal

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.types import ColumnName


class MovingAverageStep(BaseStep):
    name: Literal["movingaverage"] = "movingaverage"

    value_column: ColumnName | None = Field(None, alias="valueColumn")
    column_to_sort: ColumnName = Field(alias="columnToSort")
    moving_window: int = Field(alias="movingWindow")
    groups: list[ColumnName] = []
    new_column_name: ColumnName | None = Field(alias="newColumnName")



from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps.replacetext import ReplaceTextStep


def translate_replacetext(step: ReplaceTextStep) -> list[MongoStep]:
    return [
        {
            "$set": {
                step.search_column: {
                    "$replaceAll": {
                        "input": f"${step.search_column}",
                        "find": step.old_str,
                        "replacement": step.new_str,
                    }
                }
            }
        }
    ]
